/* $Id: mhz_capture.c,v 1.3 2013/08/29 23:44:46 jrothwei Exp jrothwei $ */
/* Joseph Rothweiler, Sensicomm LLC. Started 29Aug2013.
** Demo program to open the USB device and read samples.
*/
#include <stdio.h>
// #include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>

#include <libusb-1.0/libusb.h>

#include "usb_lib.h"
#include "mhz_lib.h"

// Catch ctrl-c to try to exit cleanly.
int do_reattach = 0;
int exitflag = 0;
void mysigint(int signum) {
	exitflag = 1;
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
static FILE *check_fopen(char *f, char *m) {
	FILE *rtn;
	rtn = fopen(f,m);
	if(rtn == NULL) {
		fprintf(stderr,"Trouble opening file <%s> mode <%s>\n",f,m);
		exit(2);
	}
	return rtn;
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
int main(int argc, char *argv[]) {
	int rtn;
	libusb_context *context;
	char *ofilename = NULL;
	int k;
	int intf_num = 0;
	int intf_alt = 0;
	int verbosity = 100;
	struct ep5_values ep5_vals;
	double cicfreq = 5e6;  // Sampling frequency in Hz after CIC.
	double ddsfreq = 10e3; // Sampling frequency in Hz after CIC.
	int posttrig = 2000;   // Stop capture posttrig samples after trigger (1-2047)

	// Process the command line.

	for(k=1;k<argc;k++) {
		if(strncmp("-d",argv[k],2)==0) {
			verbosity = strtol(argv[k]+2,0,0);
		} else if(strncmp("-o",argv[k],2)==0) {
			ofilename=argv[k]+2;
		} else if(strncmp("-f",argv[k],2)==0) {
			cicfreq = strtod(argv[k]+2,0);
		} else if(strncmp("-s",argv[k],2)==0) {
			ddsfreq = strtod(argv[k]+2,0);
		} else if(strncmp("-p",argv[k],2)==0) {
			posttrig = strtol(argv[k]+2,0,0);
		} else {
			fprintf(stderr,"Bad argument <%s>\n",argv[k]);
			break;
		};
	}
	if(k<argc) {
		fprintf(stderr,"cmd: %s [args]\n",argv[0]);
		fprintf(stderr," -d# Debug level (%d)\n",verbosity);
		fprintf(stderr," -o$ Output file name(%s)\n",ofilename);
		fprintf(stderr," -f# Sampling frequency after CIC downsampling (%g)\n",cicfreq);
		fprintf(stderr," -s# Frequency of the synthesized test signal (%g)\n",ddsfreq);
		fprintf(stderr," -p# Stop capture 1-2047 samples after trigger (%d)\n",posttrig);
		exit(2);
	}
	if(ofilename==NULL) {
		fprintf(stderr,"No output file\n");
	}

	signal(SIGINT,mysigint);

	/******************************
	** Init libusb-1.0.          */

	rtn = libusb_init (&context);
	if(rtn != 0) {
		fprintf(stderr,"%s: init trouble %d\n",argv[0],rtn);
		exit(2);
	}
	libusb_set_debug(context,3); // Info to stdout, warn/error to stderr.

	/******************************
	** Find a useful device. */

	libusb_device *found = NULL;
	found = find_usb_device(context,0x00fe,0x2100,1) ;
	if (!found) {
		fprintf(stderr,"Device not found\n");
		exit(2);
	} else {
		fprintf(stderr,"Found it\n");
	}
	// fprintf(stdout,"D\n"); fflush(stdout);

	libusb_device_handle *handle = NULL;
	// fprintf(stdout,"E\n"); fflush(stdout);
	if (found) {
		int err;

		fprintf(stderr,"Opening\n");
		err = libusb_open(found, &handle);
		if (err) {
			fprintf(stderr,"Trouble opening device %d %s\n",err,errorname(err));
			exit(2);
		}

		do_reattach = usb_device_setup(handle,intf_num,intf_alt);

		if(verbosity > 10) usb_dumpinfo(found);

		mhz_setup(handle);

		// Set the DAC frequency, so we have a test signal, if desired.
		// If not using the DAC board, the desired frequency can be zero.
		// With a 0.1uF DC blocking cap, the lower 3dB point is about 30kHz.
		// The upper limit hasn't been characterized, but a few MHz works.
		double dacfreq_actual;
		// dacfreq = mhz_setdac(handle,1e5);
		dacfreq_actual = mhz_setdac(handle,ddsfreq);
		fprintf(stdout,"DAC DDS set to %.4g MHz\n",dacfreq_actual/1e6);

		double samfreq;
		// int rtshift = 12; Clips.
		int rtshift = 13;
		samfreq = mhz_setcic(handle,cicfreq,rtshift);
		fprintf(stdout,"Sampling Frequency %.3g MHz shift %d\n",samfreq,rtshift);

		// Reset the limit monitors and start capture
		mhz_regset(handle,MHZREG_STROBEBITS,MHZBITS_HILORESET | MHZBITS_CAPSTART);

		mhz_regset(handle,MHZREG_POSTTRIG,posttrig);
		mhz_trigger(handle,MHZTRIG_EXTERNAL,0); 

		ep5_vals.cicwrit = 1; // Use this value before the first call to ep5_unpack.
		int loopcount;
		for(loopcount=0;!exitflag;loopcount++) {
			int8_t indata[512];
			int timeout_ms = 100;
			int actlen;

			// If a capture has been completed, grab the data.
			if( (ep5_vals.cicwrit & 1) == 0) {
				FILE *outfile;
				outfile = check_fopen(ofilename,"w");
				mhz_get_the_data(handle,outfile,3,&ep5_vals);
				fclose(outfile);
				fprintf(stdout,"Data written to %s\n",ofilename);
				break;
			}
			// Main data channel. This never executes unless there was a problem,
			// such as leftover data from a previous run of the program.
			// See mhz_get_the_data for the real read.
			checked_bulk_transfer (handle,0x83,(char *)indata,512,&actlen,timeout_ms,1);
			if(actlen>0) {
				fprintf(stdout,"Data in %d\n",actlen);
			}
			// Look for a status message on the interrupt channel. This happens for
			// one of two reasons:
			// - A status check was requested with the MHZBITS_INT_NOW strobe
			// - A capture has been completed. The status messsage is automatically
			//   requested by the VHDL code in this case.
			checked_bulk_transfer (handle,0x85,(char *)indata,512,&actlen,timeout_ms,0);
			if(actlen>0) {
				// fprintf(stdout,"Chan 5 in %d\n",actlen);
				ep5_unpack(indata,actlen,&ep5_vals);
				fprintf(stdout,"Loop %d: ",loopcount);
				ep5_display(stdout,&ep5_vals);
			}

			// Trigger a status check. Results in a channel 5 message. This is optional.
			if((loopcount%10)==0) mhz_regset(handle,MHZREG_STROBEBITS,MHZBITS_INT_NOW);
		}
	} else {
		fprintf(stderr,"Not found.\n");
	}
	if(do_reattach) {
		int err;
		err = libusb_attach_kernel_driver(handle,intf_num);
		if(err == 0) {
			fprintf(stderr,"kernel driver reattached\n");
		} else {
			fprintf(stderr,"Error %d %s on kernel driver reattach\n",err,errorname(err));
		}
	} else {
		fprintf(stderr,"No driver to reattach\n");
	}

	/******************************
	** Done. */
	if(handle != NULL) libusb_close(handle);
	libusb_exit (context);
	return 0;
}
