/* $Id: mhz_lib.c,v 1.3 2013/08/29 23:45:55 jrothwei Exp jrothwei $ */
/* Joseph Rothweiler, Sensicomm LLC. Started 29Aug2013
** Subroutines to simplify operation of the MHZ100Q2 board with oc_scope firmware.
**
** Endpoints used are:
** 0x83 - Input main data channel.
** 0x04 - Output control channel for command characters.
** 0x85 - Input status interrupt channel.
**
** Command characters for EP 5 are:
** 0-9,a-f - hex bytes
** : - Clear accum
** @ - Copy accum to addr
** = - Copy accum to data
**
** Addresses are:
** 0 - Read address
** 1 - Read length
** 2 - Decimation
** 3 - right shift.
** 4 - dac8step
** 6 - posttrigger
** 8 - Clear selected control bits.
** 9 - Set selected control bits.
** a - strobe lines
**
** Control bits for 8 and 9 are:
** 0 trigger_enable
** 1 trigger_on_rising
** 2 trigger_extern
** 3 channel select (for read)
** 4 pip (old debug thing)

** Strobe bits for address 10 (0xa) are:
** chanx_hilo_reset    0
** ep5_trigger_s48     1
** trigger_manual      2
** chanx_cic_write_set 3 (capture start)
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

#include <libusb-1.0/libusb.h>

#include "usb_lib.h"

#define ArraySize(x) (sizeof(x)/sizeof(*x))

/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
static uint64_t llpackvals(int8_t *data, int len, int beg, int end) {
	uint64_t rtn;
	int k;
	rtn = 0;
	if(end>=len) return rtn;
	for(k=beg;k<=end;k++) rtn = (rtn<<8) | ((uint8_t *)data)[k];
	return rtn;
}
static uint32_t packvals(int8_t *data, int len, int beg, int end) {
	return (uint32_t)llpackvals(data, len, beg, end);
}

/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
/*Hstart*/
enum mhz_vhdl_regs {
	MHZREG_READADDR   = 0,
	MHZREG_READLEN    = 1,
	MHZREG_CIC        = 2,
	MHZREG_RTSHIFT    = 3,
	MHZREG_DACDDS     = 4,
	MHZREG_POSTTRIG   = 6,
	MHZREG_CLEARBITS  = 8,
	MHZREG_SETBITS    = 9,
	MHZREG_STROBEBITS = 0xa
};
enum mhz_trigger_modes {
	MHZTRIG_OFF      ,
	MHZTRIG_RISING   ,
	MHZTRIG_FALLING  ,
	MHZTRIG_EXTERNAL ,
	MHZTRIG_NOW
} ;
#define MHZEP_COMMAND (0x4)
// For the strobes.
#define MHZBITS_HILORESET (0x1)
#define MHZBITS_INT_NOW   (0x2)
#define MHZBITS_TRIG_NOW  (0x4)
#define MHZBITS_CAPSTART  (0x8)
// For the settings.
#define MHZBITS_TRIG_ENABLE (0x1)
#define MHZBITS_TRIG_RISING (0x2)
#define MHZBITS_TRIG_EXTERN (0x4)
#define MHZBITS_CHAN_SELECT (0x8)
#define MHZBITS_ALLCTL      (0x1f)
void mhz_regset(libusb_device_handle *handle, uint32_t reg, uint32_t val)
/*Hend;*/ {
	char tmpstr[22];
	int rtn;
	rtn = snprintf(tmpstr,ArraySize(tmpstr),":%x@:%x=",reg,val);
	if(rtn >= ArraySize(tmpstr) ) {
		// This should never happen.
		fprintf(stderr,"Internal error in mhz_regset %x %x -> %s\n",reg,val,tmpstr);
		exit(2);
	}
	int actlen;
	int timeout_ms = 100;
	checked_bulk_transfer (handle,MHZEP_COMMAND,
		tmpstr,strlen(tmpstr),
		&actlen,timeout_ms,
		0x1    // Print errors only.
	);
}
/*Hstart*/ void mhz_setup(
	libusb_device_handle *handle
)
/*Hend;*/ {
	// Clear all the control bits.
	// FIXME: This doesn't fully reset everything: If a capture is in progress
	// it will run to completion.
	mhz_regset(handle,MHZREG_CLEARBITS,MHZBITS_ALLCTL);
}
/*Hstart*/ double mhz_setdac(
	libusb_device_handle *handle,
	double freq
)
/*Hend;*/ {
	/* The internal synthesizer is a 32-bit counter driven at 100MHz
	** The upper bits address a sine table which then drives the DAC.
	** So one output cycle is 2**32 counts and the frequency is therefore
	** freq = increment * 100e6/(2**32); */
	double rtn;
	uint32_t stepvalue;
	double val232 = (double)0x10000*(double)0x10000; // 2**32
	double factor = 100e6/val232; // 100 MHz / 2**32.
	double dblstep = freq/factor;
	if(dblstep < 0) dblstep = 0;
	if(dblstep > (val232-1.0) ) dblstep = val232-1.0;
	stepvalue = dblstep;
	rtn = factor*(double)stepvalue;
	mhz_regset(handle,MHZREG_DACDDS,stepvalue);
	return rtn;
}
/*Hstart*/ double mhz_setcic( // Returns actual frequency.
	libusb_device_handle *handle,
	double freq, // Desired sampling frequency.
	int bitshift
)
/*Hend;*/ {
	double factor;
	if(freq > 0) factor = 100e6/freq;
	else         factor = 1.0;
	if(factor<1.0) factor = 1.0;
	if(factor>100.0) factor = 100.0;
	int setval = factor-0.5; // Truncate, so effectively setval=round(factor-1).
	mhz_regset(handle,MHZREG_CIC,setval);

	// Also, set the right-shift value.
	if(bitshift<0) bitshift = 0;
	if(bitshift>24) bitshift = 24;
	mhz_regset(handle,MHZREG_RTSHIFT,bitshift);
	return 100e6/(setval+1);
}
/*Hstart*/ void mhz_trigger(
	libusb_device_handle *handle,
	enum mhz_trigger_modes trigmode,
	int triggerlevel
)
/*Hend;*/ {
	switch(trigmode) {
	case (MHZTRIG_OFF) :
		// Clear the trigger flags.
		mhz_regset(handle,MHZREG_CLEARBITS,
			MHZBITS_TRIG_ENABLE | MHZBITS_TRIG_RISING | MHZBITS_TRIG_EXTERN );
		break;
	case (MHZTRIG_RISING) :
		// Clear all, set level, then  set enable and rising.
		mhz_regset(handle,MHZREG_CLEARBITS,
			MHZBITS_TRIG_ENABLE | MHZBITS_TRIG_RISING | MHZBITS_TRIG_EXTERN );
		// FIXME: Oops: Trigger level is always zero in the current VHDL firmware.
		mhz_regset(handle,MHZREG_SETBITS,
			MHZBITS_TRIG_ENABLE | MHZBITS_TRIG_RISING);
		break;
	case (MHZTRIG_FALLING) :
		// Clear all, set level, then  set enable.
		mhz_regset(handle,MHZREG_CLEARBITS,
			MHZBITS_TRIG_ENABLE | MHZBITS_TRIG_RISING | MHZBITS_TRIG_EXTERN );
		// FIXME: Oops: Trigger level is always zero in the current VHDL firmware.
		mhz_regset(handle,MHZREG_SETBITS,
			MHZBITS_TRIG_ENABLE);
		break;
	case (MHZTRIG_EXTERNAL) :
		// Clear internal trigger.
		mhz_regset(handle,MHZREG_CLEARBITS,
			MHZBITS_TRIG_ENABLE | MHZBITS_TRIG_RISING);
		// Set external. No harm done if it's already set.
		mhz_regset(handle,MHZREG_SETBITS,
			MHZBITS_TRIG_EXTERN);
		break;
	case (MHZTRIG_NOW) :
		// Assume trigger is already set appropriately.
		// Just strobe to force a capture.
		mhz_regset(handle,MHZREG_STROBEBITS, MHZBITS_TRIG_NOW) ;
		break;
	default:
		// All enums covered, so this should never happen.
		fprintf(stderr,"Internal error in mhz_trigger\n");
		exit(2);
		break;
	}
}
/*Hstart*/ struct ep5_values {
        uint32_t r1      ;
        uint32_t r2      ;
        uint32_t zip1    ;
        uint32_t posttrg ;
        uint32_t cicwrit ;
        uint32_t capcnt  ;
        uint32_t wr_addr ;
        uint32_t cmddata ;
        int32_t  adc0max ;
        int32_t  adc0min ;
        uint64_t cic0min ;
        uint64_t cic0max ;
};
void ep5_unpack(int8_t *data, int len, struct ep5_values *vals)
/*Hend;*/ {
	// Breakouts, updated for rev 1.18.
	vals->r1      =          packvals(data,len, 0, 3);
	vals->r2      =          packvals(data,len, 4, 7);
	vals->zip1    =          packvals(data,len, 8,11);
	vals->posttrg =          packvals(data,len,12,13)>>4;
	vals->cicwrit =          packvals(data,len,13,13)&0xf; // Really 1 bit.
	vals->capcnt  =          packvals(data,len,14,14)>>4;
	vals->wr_addr =          packvals(data,len,14,15)&0xfff;
	vals->cmddata =          packvals(data,len,16,19);
	vals->adc0max = (int32_t)packvals(data,len,20,20);
	vals->adc0min = (int32_t)packvals(data,len,21,21);
	vals->cic0min =          llpackvals(data,len,22,26);
	vals->cic0max =          llpackvals(data,len,27,31);
}
/*Hstart*/ void ep5_display(FILE *outf, struct ep5_values *vals)
/*Hend;*/ {
	fprintf(outf," post %3x cic %x cap %x wr %3x cmd %8x adc %2x %2x cic %10lx %10lx",
		vals->posttrg,
		vals->cicwrit,
		vals->capcnt,
		vals->wr_addr,
		vals->cmddata,
		vals->adc0min,
		vals->adc0max,
		vals->cic0min,
		vals->cic0max
	);
	/* These are sanity checks. r1 and r2 always have these values for now. */
	if(vals->r1!=0x01234567) fprintf(outf," r1 0x%x",vals->r1);
	if(vals->r2!=0x89abcdef) fprintf(outf," r2 0x%x",vals->r2);
	/* zip1 is for VHDL debug. */
	if(vals->zip1!=0) fprintf(outf," zip1 0x%x",vals->zip1);
	fprintf(outf,"\n");
}
//------------------------------------------------------------------------
static void multiread(libusb_device_handle *handle, int8_t *buf, int len) {
	int timeout_ms = 100;
	int n;
	int offset;
	offset = 0;
	for( ; len> 0; ) {
		int actlen;
		n = len;
		if(n > 256) n = 256;
		mhz_regset(handle, MHZREG_READADDR,offset);
		mhz_regset(handle, MHZREG_READLEN,n); // This triggers the upload.
 		checked_bulk_transfer (handle,0x83,(char *)buf,n,&actlen,timeout_ms,1);
		if(actlen != n) {
			fprintf(stderr,"Oops: mhz_get_the_data requested %d , got %d\n", n, actlen);
			exit(2);
		}
		buf += actlen;
		offset += actlen;
		len -= actlen;
	}
}
/*Hstart*/ void mhz_get_the_data(
	libusb_device_handle *handle,
	FILE *outfile,
	int chanmask,
	struct ep5_values *vals
)
/*Hend;*/ {
	int bufferlen = 2048;
	int8_t chan0[bufferlen];
	int8_t chan1[bufferlen];
	// Save the status info.
	fprintf(outfile,"INT ");
	ep5_display(outfile,vals);
	fprintf(outfile,"MASK %x\n",chanmask);

	// Upload the data.
	// FIXME: VHDL glitch: we want to read 2048 values, but the register
	// is 11 bits, which means that 2048==0. So we do the read in two chunks.
	if(chanmask&1) { // Channel 0
		// Clear MHZBITS_CHAN_SELECT to select channel 0.
		mhz_regset(handle,MHZREG_CLEARBITS, MHZBITS_CHAN_SELECT);
		// Now, get the data.
		multiread(handle,chan0,bufferlen);
	}
	if(chanmask&2) { // Channel 1
		// Set MHZBITS_CHAN_SELECT to select channel 1.
		mhz_regset(handle,MHZREG_SETBITS, MHZBITS_CHAN_SELECT);
		// Now, get the data.
		multiread(handle,chan1,bufferlen);
	}

	// Now, save the data to the buffer. The RAM in the FPGA is used as a circular buffer, so
	// unwrap it here.
	int ptr = vals->wr_addr;
	int k;
	if((chanmask&3)!=0) for(k=0;k<bufferlen;k++) {
		ptr = (ptr+1)%bufferlen;
		if(chanmask&1) fprintf(outfile," %4d",(int)chan0[ptr]);
		if(chanmask&2) fprintf(outfile," %4d",(int)chan1[ptr]);
		fprintf(outfile,"\n");
	}
}
