/* $Id: usb_lib.c,v 1.2 2013/08/29 14:36:16 jrothwei Exp $ */
/* Joseph Rothweiler, Sensicomm LLC. Extracted from usbio.c on 19Aug2013.
** routines for use with libusb 1.0 .
*/

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>        /* For mode constants */
#include <fcntl.h>           /* For O_* constants */
#include <time.h>
#include <mqueue.h>
#include <ctype.h>

#include <libusb-1.0/libusb.h>

#define ArraySize(x) (sizeof(x)/sizeof(*x))

//-----------------------------------------------------------------------------
// Utility routines.
static void printit(char *buf, int len) {
	int k;
	for(k=0;k<len;k++) {
		if(k%16==0) fprintf(stderr,"%4x:",k);
		fprintf(stderr," %2.2x",buf[k]);
		if(k%16==15) fprintf(stderr,"\n");
	}
	if(k%16!=0) fprintf(stderr,"\n");
}
//--------------
static void printchars(char *buf, int len) {
	int k;
	fprintf(stderr,"string[%d] = \"",len);
	for(k=0;k<len;k++) {
		if(buf[k]=='\\') fprintf(stderr,"\\\\");
		else if(isprint(buf[k])) fprintf(stderr,"%c",buf[k]);
		else fprintf(stderr,"\\%hhx",buf[k]);
	}
	fprintf(stderr,"\"\n");
}
//--------------
static void smalldump(unsigned char *buf, int n) {
	int k;
	if(n>32) n = 32;
	for(k=0;k<n;k++) fprintf(stdout," %2.2x",buf[k]);
	fprintf(stdout,"\n");
}
//-----------------------------------------------------------------------------

// From the libusb include files. To turn error returns into strings.
static struct errorname_table_struct {
  char *name;
  int val;
} errorname_table[] = {
  { "SUCCESS",               0 },
  { "ERROR_IO",             -1 },
  { "ERROR_INVALID_PARAM",  -2 },
  { "ERROR_ACCESS",         -3 },
  { "ERROR_NO_DEVICE",      -4 },
  { "ERROR_NOT_FOUND",      -5 },
  { "ERROR_BUSY",           -6 },
  { "ERROR_TIMEOUT",        -7 },
  { "ERROR_OVERFLOW",       -8 },
  { "ERROR_PIPE",           -9 },
  { "ERROR_INTERRUPTED",   -10 },
  { "ERROR_NO_MEM",        -11 },
  { "ERROR_NOT_SUPPORTED", -12 },
  { "ERROR_OTHER",         -99 },
  { "Undefined",           -99 } // Name if none of the others are found.
};
//-------------------
/*Hstart*/ const char *errorname(int err)
/*Hend;*/ {
	int k;
	int tablesize = sizeof(errorname_table)/sizeof(errorname_table[0]);
	for(k=0;k<tablesize-1;k++) {
		if(err == errorname_table[k].val) break;
	}
	return errorname_table[k].name;
}
//-------------------
static void errdump(char *id, int err, int the_errno) {
	fprintf(stderr,"Error at %s : %d -> %s ; errno %d -> %s\n",
		id, err, errorname(err),the_errno,strerror(the_errno) );
}
//------------------------------------------------------------------------------
// Input or output transfer with libusb status checking.
/*Hstart*/ void checked_bulk_transfer (
  struct libusb_device_handle *handle,
  unsigned char  endpoint,  // >= 0x80 for IN transfer.
  char * data,
  int len,
  int *actlen,
  unsigned int timeout_ms,
  int verbose // 1: errors, 2: nonzero xfers, 4: all xfers, 8:timeout, 0x10: payload 0x20: chars
)
/*Hend;*/ {
	int err;
	int the_errno;
	*actlen = 0;
	err = libusb_bulk_transfer(handle,endpoint,(unsigned char *)data,len,actlen,timeout_ms);
	the_errno = errno;
	if(err == 0) {
		if((verbose&2)&&(*actlen>0)) fprintf(stderr,"EP %2.2x Transferred %d\n",endpoint&0xff,*actlen);
		else if(verbose&4) fprintf(stderr,"EP %2.2x Transferred %d\n",endpoint&0xff,*actlen);
		if(verbose&0x10) printit(data,*actlen);
		if(verbose&0x20) printchars(data,*actlen);
	} else if(err == LIBUSB_ERROR_TIMEOUT) {
		if(verbose&8) fprintf(stderr,"EP %2.2x Timeout %d\n",endpoint&0xff,*actlen);
		/* Do nothing for now. */
	} else if(verbose&1) {
		fprintf(stderr,"EP %2.2x Error %d Transferred %d\n",endpoint&0xff,err,*actlen);
		fprintf(stderr,"errno %d -> %s\n",the_errno,strerror(the_errno) );
	}
}
// Continuous transfer to a file. 
/*Hstart*/ void continuous_transfer(
	struct libusb_device_handle *handle,
	int ofile,
	unsigned char endpoint,
	volatile int *exitflag // Set nonzero to end transfer.
)
/*Hend;*/ {
	while( !(*exitflag) ) {
		unsigned char buf[512];
		int buflen = 512;
		int actlen;
		int err;
		int the_errno;
		int timeout_ms=10;
		actlen = 0;
		err = libusb_bulk_transfer(handle,endpoint,buf,buflen,&actlen,timeout_ms);
		the_errno = errno;
		if(err == 0) {
			fprintf(stderr,"EP %2.2x Transferred %d\n",endpoint&0xff,actlen);
			smalldump(buf,actlen);
			if((actlen>0)&&(ofile>=0)) {
				ssize_t rtn;
				rtn = write(ofile,buf,actlen);
				if(rtn != actlen) {
					fprintf(stderr,"Trouble writing output file\n");
					break;
				}
			}
		} else if(err == LIBUSB_ERROR_TIMEOUT) {
			fprintf(stderr,"EP %2.2x timeout. Count %d\n",(int)endpoint,actlen);
			smalldump(buf,actlen);
			// Do nothing for now.
		} else {
			fprintf(stderr,"Testing A\n");
			fprintf(stderr,"EP %2.2x Error %d %s Transferred %d\n",endpoint&0xff,err,errorname(err),actlen);
			fprintf(stderr,"Testing B\n"); fflush(stderr);
			fprintf(stderr,"errno %d -> %s\n",the_errno,strerror(the_errno) ); fflush(stderr);
			fprintf(stderr,"Testing C\n"); fflush(stderr);
			break;
		}
		// if(actlen>0) printit(buf,actlen);
	}
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
// Helper routine to print USB errors.
/*Hstart*/ void usberrdump(char *id, int err, int the_errno)
/*Hend;*/ {
	fprintf(stderr,"Error at %s : %d -> %s ; errno %d -> %s\n",
		id, err, errorname(err),the_errno,strerror(the_errno) );
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
// For diagnostics.
static void dumpdevdesc( struct libusb_device_descriptor *desc) {
	fprintf(stderr,"Device Descriptor\n");
	fprintf(stderr," bLength 0x%x type 0x%x USB ver 0x%x\n",
	0 + desc -> bLength,
	0 + desc -> bDescriptorType,
	0 + desc -> 	bcdUSB );
	fprintf(stderr," Class 0x%x SubClass 0x%x Protocol 0x%x\n",
	0 + desc -> 	bDeviceClass,
	0 + desc -> 	bDeviceSubClass,
	0 + desc -> 	bDeviceProtocol );
	fprintf(stderr," MaxPacketSize0 0x%x\n",
	0 + desc -> 	bMaxPacketSize0 );
	fprintf(stderr," Vendor 0x%x Product 0x%x Device 0x%x\n",
	0 + desc -> 	idVendor,
	0 + desc -> 	idProduct,
	0 + desc -> 	bcdDevice );
	fprintf(stderr," iMfr 0x%x iProd 0x%x iSerNo 0x%x\n",
	0 + desc -> 	iManufacturer ,
	0 + desc -> 	iProduct ,
	0 + desc -> 	iSerialNumber );
	fprintf(stderr," NumConfig 0x%x\n",
	0 + desc -> 	bNumConfigurations );
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
static void decode_attribs(uint8_t attribs) {
	uint8_t v1;
	v1 = attribs&3;
	fprintf(stderr,"Attribs 0x%2.2x :",0+attribs);
		fprintf(stderr," Type ");
	switch(v1) {
	case LIBUSB_TRANSFER_TYPE_CONTROL :
		fprintf(stderr,"CTRL");
		break;
	case LIBUSB_TRANSFER_TYPE_ISOCHRONOUS :
		fprintf(stderr,"ISOC");
		break;
	case LIBUSB_TRANSFER_TYPE_BULK :
		fprintf(stderr,"BULK");
		break;
	case LIBUSB_TRANSFER_TYPE_INTERRUPT :
		fprintf(stderr,"INTR");
		break;
	default:
		fprintf(stderr,"????");
		break;
	}
	fprintf(stderr," Sync ");
	v1 = attribs&0xc;
	switch(v1) {
	case LIBUSB_ISO_SYNC_TYPE_NONE :
		fprintf(stderr,"NONE");
		break;
	case LIBUSB_ISO_SYNC_TYPE_ASYNC :
		fprintf(stderr,"ASYN");
		break;
	case LIBUSB_ISO_SYNC_TYPE_ADAPTIVE :
		fprintf(stderr,"ADAP");
		break;
	case LIBUSB_ISO_SYNC_TYPE_SYNC :
		fprintf(stderr,"SYNC");
		break;
	default:
		fprintf(stderr,"????");
		break;
	}
	fprintf(stderr," Usage ");
	v1 = attribs&0x30;
	switch(v1) {
	case LIBUSB_ISO_USAGE_TYPE_DATA :
		fprintf(stderr,"DATA");
		break;
	case LIBUSB_ISO_USAGE_TYPE_FEEDBACK :
		fprintf(stderr,"FDBK");
		break;
	case LIBUSB_ISO_USAGE_TYPE_IMPLICIT :
		fprintf(stderr,"IMPL");
		break;
	default:
		fprintf(stderr,"????");
		break;
	}
	fprintf(stderr," Reserved ");
	v1 = attribs&0xc0;
	fprintf(stderr,"0x%2.2x",0+v1);

	fprintf(stderr,"\n");
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
static void dumpepdesc(const struct libusb_endpoint_descriptor *endp, int num) {
	static char dir_in[] ="IN";
	static char dir_out[]="OUT";
	static char dir_bad[]="BAD";
	static char *dir_val;
	if( (0x80 & endp->bEndpointAddress) == LIBUSB_ENDPOINT_IN) dir_val = dir_in;
	else if( (0x80 & endp->bEndpointAddress) == LIBUSB_ENDPOINT_OUT) dir_val = dir_out;
	else dir_val = dir_bad;

	fprintf(stderr,"Endpoint %d Len 0x%x DescType 0x%x\n",
	num,
	0 + endp->bLength,
	0 + endp->bDescriptorType );
	fprintf(stderr,"Addr 0x%x %3s Attrib 0x%x PktSiz 0x%x\n",
	0 + endp->bEndpointAddress, dir_val,
	0 + endp->bmAttributes,
	0 + endp->wMaxPacketSize );
	fprintf(stderr,"Interval 0x%x Refresh 0x%x SyncAccr 0x%x\n",
	0 + endp->bInterval,
	0 + endp->bRefresh,
	0 + endp->bSynchAddress );
	// const unsigned char *extra
	fprintf(stderr,"Extra Len 0x%x\n",
	0 + endp->extra_length );
	decode_attribs(endp->bmAttributes);
	if(endp->extra_length != 0) {
		fprintf(stderr,"Extra TBD\n");
	}
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
static void dumpinterfacedesc(const struct libusb_interface_descriptor * ifp, int num, int alt ) {
	fprintf(stderr,"Interface %d Alt %d Len 0x%x Type 0x%x Number 0x%x Alt 0x%x\n",
	num, alt,
	0 + ifp->bLength,
	0 + ifp->bDescriptorType,
	0 + ifp->bInterfaceNumber,
	0 + ifp->bAlternateSetting );
	fprintf(stderr,"Endpoints 0x%x Class 0x%x SubClass 0x%x Protocol 0x%x\n",
	0 + ifp->bNumEndpoints,
	0 + ifp->bInterfaceClass,
	0 + ifp->bInterfaceSubClass,
	0 + ifp->bInterfaceProtocol );
	fprintf(stderr,"iInterface 0x%x extra_len 0x%x\n",
	0 + ifp->iInterface,
	0 + ifp->extra_length );
	int k;
	for(k=0;k<ifp->bNumEndpoints;k++) {
		dumpepdesc(ifp->endpoint+k,k);
	}
	if(ifp->extra_length != 0) {
		fprintf(stderr,"Extra TBD\n");
		// const unsigned char * 	extra
	}
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
static void dumpconfigdesc( struct libusb_config_descriptor *cfg, int num) {
	fprintf(stderr,"Config %d descriptor. Length 0x%x Type 0x%x Total Len 0x%x\n",
	num,
	0 + cfg->bLength,
	0 + cfg->bDescriptorType,
	0 + cfg->wTotalLength );
	fprintf(stderr,"NumInt 0x%x ConfigValue 0x%x iConfig 0x%x Attribs 0x%x MaxPwr 0x%x\n",
	0 + cfg->bNumInterfaces,
	0 + cfg->bConfigurationValue,
	0 + cfg->iConfiguration,
	0 + cfg->bmAttributes,
	0 + cfg->MaxPower );
	fprintf(stderr,"Extra Len 0x%x\n",
	0 + cfg->extra_length );
	int k;
	for(k=0;k<cfg->bNumInterfaces;k++) {
		const struct libusb_interface_descriptor * ifp;
		int p;
		p = cfg->interface[k].num_altsetting;
		ifp = cfg->interface[k].altsetting;
		dumpinterfacedesc(ifp,k,p);
	}
	if(cfg->extra_length != 0 ) {
		fprintf(stderr,"Extra TBD\n");
		// const unsigned char * 	extra
	}
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
/*Hstart*/ void usb_dumpinfo(libusb_device *usbdev)
/*Hend;*/ {
	int err;
	int k;
	struct libusb_device_descriptor desc ;
	err = libusb_get_device_descriptor (usbdev, &desc);
	if(err != 0) errdump("dumpinfo_descriptor",err,errno);
	else dumpdevdesc(&desc);

	for(k=0;k<desc.bNumConfigurations;k++) {
		struct libusb_config_descriptor *config;
		err = libusb_get_config_descriptor (usbdev, k, &config);
		if(err != 0) errdump("dumpinfo_config",err,errno);
		else dumpconfigdesc(config,k);
	}

}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
// If chk is true exit on error, else return NULL.
/*Hstart*/ libusb_device *find_usb_device(libusb_context *context, uint16_t vend, uint16_t prod, int chk)
/*Hend;*/ {
	/******************************
	** Find a useful device. */
	libusb_device **usbdevlist;
	libusb_device *found = NULL;
	// fprintf(stdout,"F1\n"); fflush(stdout);
	ssize_t cnt = libusb_get_device_list(context, &usbdevlist);
	ssize_t i = 0;
	if (cnt < 0) {
		fprintf(stderr,"Trouble get_device_list %ld\n",cnt);
		if(chk) exit(2);
		else return found;
	}

	for (i = 0; i < cnt; i++) {
		fprintf(stderr,"Checking device %ld of %ld\n",i,cnt);
		// libusb_device *device = usbdevlist[i];
		struct libusb_device_descriptor desc;
		int r = libusb_get_device_descriptor(usbdevlist[i], &desc);
		if (r < 0) {
			fprintf(stderr, "failed to get device descriptor");
			if(chk) exit(2);
			else return found;
		} else {
			fprintf(stderr,"%04x:%04x (bus %d, device %d)\n",
				desc.idVendor, desc.idProduct,
				libusb_get_bus_number(usbdevlist[i]),
				libusb_get_device_address(usbdevlist[i]));
		}
		if ( (desc.idVendor==vend) && (desc.idProduct==prod) ) {
			found = usbdevlist[i];
			break;
		}
	}
	if(!found) {
		fprintf(stderr,"Device vendor 0x%4.4hx prod 0x%4.4hx not found\n",vend,prod);
		if(chk) exit(2);
	}
	// libusb_free_device_list(usbdevlist, 1);
	return found;
}
//-----------------------------------------------------------------------------
/*Hstart*/ int usb_device_setup(
	struct libusb_device_handle *handle,
	int intf_num,
	int intf_alt
)
/*Hend;*/ {
	int err;
	int do_reattach;
	do_reattach = 0;
	err = libusb_kernel_driver_active(handle,intf_num);
	if(err == 0) {
		fprintf(stderr,"No kernel driver active.\n");
	} else if(err == 1) {
		fprintf(stderr,"Releasing kernel driver.\n");
		err = libusb_detach_kernel_driver(handle,intf_num);
		if(err == 0) {
			fprintf(stderr,"kernel driver successfully detached\n");
			do_reattach = 1;
		} else {
			fprintf(stderr,"Error code %d %s detaching driver\n",err,errorname(err));
			exit(2);
		}
	} else {
		fprintf(stderr,"error %d %s from libusb_kernel_driver_active\n",err,errorname(err));
		exit(2);
	}
	err = libusb_claim_interface(handle,intf_num);
	if (err) {
		fprintf(stderr,"Trouble claim_interface %d %s\n",err,errorname(err));
		exit(2);
	}
	err = libusb_set_interface_alt_setting (handle,intf_num,intf_alt);
	if (err) {
		fprintf(stderr,"Trouble alt_setting %d %s\n",err,errorname(err));
		exit(2);
	}
	return do_reattach;
}
