/* $Id: usbio.c,v 1.12 2013/08/29 23:58:13 jrothwei Exp jrothwei $ */
/* Joseph Rothweiler, Sensicomm LLC. Started 21Sep2011.
** Access the USB device and save the resulting data in shared memory.
** Another process will display it.
*/

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>        /* For mode constants */
#include <fcntl.h>           /* For O_* constants */
#include <time.h>
#include <mqueue.h>
#include <ctype.h>

#include <libusb-1.0/libusb.h>

#include "shmem_layout.h"

static void printit(char *buf, int len) ;

#define ArraySize(x) (sizeof(x)/sizeof(*x))

static struct shmem_struct *sh_ptr = NULL;
// sh_ptr->vals[*] follow the VHDL code's internal register numbering.

enum vhdl_regs {
  VREG_READADDR=0 , VREG_READLEFT=1, VREG_CIC     =2, VREG_RTSHIFT=3,
  VREG_DDS     =4,  VREG_FLAGS   =5, VREG_POSTTRIG=6
};

static struct errorname_table_struct {
  char *name;
  int val;
} errorname_table[] = {
  { "SUCCESS",               0 },
  { "ERROR_IO",             -1 },
  { "ERROR_INVALID_PARAM",  -2 },
  { "ERROR_ACCESS",         -3 },
  { "ERROR_NO_DEVICE",      -4 },
  { "ERROR_NOT_FOUND",      -5 },
  { "ERROR_BUSY",           -6 },
  { "ERROR_TIMEOUT",        -7 },
  { "ERROR_OVERFLOW",       -8 },
  { "ERROR_PIPE",           -9 },
  { "ERROR_INTERRUPTED",   -10 },
  { "ERROR_NO_MEM",        -11 },
  { "ERROR_NOT_SUPPORTED", -12 },
  { "ERROR_OTHER",         -99 },
  { "Undefined",           -99 }
};
static char *errorname(int err) {
	int k;
	int tablesize = sizeof(errorname_table)/sizeof(errorname_table[0]);
	for(k=0;k<tablesize-1;k++) {
		if(err == errorname_table[k].val) break;
	}
	return errorname_table[k].name;
}
static void printchars(char *buf, int len) {
	int k;
	fprintf(stderr,"string[%d] = \"",len);
	for(k=0;k<len;k++) {
		if(buf[k]=='\\') fprintf(stderr,"\\\\");
		else if(isprint(buf[k])) fprintf(stderr,"%c",buf[k]);
		else fprintf(stderr,"\\%hhx",buf[k]);
	}
	fprintf(stderr,"\"\n");
}
static void checked_bulk_transfer (
  struct libusb_device_handle *handle,
  unsigned char  endpoint,  // >= 0x80 for IN transfer.
  char * data,
  int len,
  int *actlen,
  unsigned int timeout_ms,
  int verbose // 1: errors, 2: nonzero xfers, 4: all xfers, 8:timeout, 0x10: payload 0x20: chars
) {
	int err;
	int the_errno;
	*actlen = 0;
	err = libusb_bulk_transfer(handle,endpoint,(unsigned char *)data,len,actlen,timeout_ms);
	the_errno = errno;
	if(err == 0) {
		if((verbose&2)&&(*actlen>0)) fprintf(stderr,"EP %2.2x Transferred %d\n",endpoint&0xff,*actlen);
		else if(verbose&4) fprintf(stderr,"EP %2.2x Transferred %d\n",endpoint&0xff,*actlen);
		if(verbose&0x10) printit(data,*actlen);
		if(verbose&0x20) printchars(data,*actlen);
	} else if(err == LIBUSB_ERROR_TIMEOUT) {
		if(verbose&8) fprintf(stderr,"EP %2.2x Timeout %d\n",endpoint&0xff,*actlen);
		/* Do nothing for now. */
	} else if(verbose&1) {
		fprintf(stderr,"EP %2.2x Error %d Transferred %d\n",endpoint&0xff,err,*actlen);
		fprintf(stderr,"errno %d -> %s\n",the_errno,strerror(the_errno) );
	}
}

int do_reattach = 0;

static int exitflag = 0;
void mysigint(int signum) {
	exitflag = 1;
}
static void printit(char *buf, int len) {
	int k;
	for(k=0;k<len;k++) {
		if(k%16==0) fprintf(stderr,"%4x:",k);
		fprintf(stderr," %2.2x",buf[k]);
		if(k%16==15) fprintf(stderr,"\n");
	}
	if(k%16!=0) fprintf(stderr,"\n");
}
static void smalldump(unsigned char *buf, int n) {
	int k;
	if(n>32) n = 32;
	for(k=0;k<n;k++) fprintf(stdout," %2.2x",buf[k]);
	fprintf(stdout,"\n");
}
static void continuous_transfer(struct libusb_device_handle *handle,int ofile, unsigned char endpoint) {
	while(!exitflag) {
		unsigned char buf[512];
		int buflen = 512;
		int actlen;
		int err;
		int the_errno;
		int timeout_ms=10;
		actlen = 0;
		err = libusb_bulk_transfer(handle,endpoint,buf,buflen,&actlen,timeout_ms);
		the_errno = errno;
		if(err == 0) {
			fprintf(stderr,"EP %2.2x Transferred %d\n",endpoint&0xff,actlen);
			smalldump(buf,actlen);
			if((actlen>0)&&(ofile>=0)) {
				ssize_t rtn;
				rtn = write(ofile,buf,actlen);
				if(rtn != actlen) {
					fprintf(stderr,"Trouble writing output file\n");
					break;
				}
			}
		} else if(err == LIBUSB_ERROR_TIMEOUT) {
			fprintf(stderr,"EP %2.2x timeout. Count %d\n",(int)endpoint,actlen);
			smalldump(buf,actlen);
			// Do nothing for now.
		} else {
			fprintf(stderr,"Testing A\n");
			fprintf(stderr,"EP %2.2x Error %d %s Transferred %d\n",endpoint&0xff,err,errorname(err),actlen);
			fprintf(stderr,"Testing B\n"); fflush(stderr);
			fprintf(stderr,"errno %d -> %s\n",the_errno,strerror(the_errno) ); fflush(stderr);
			fprintf(stderr,"Testing C\n"); fflush(stderr);
			break;
		}
		// if(actlen>0) printit(buf,actlen);
	}
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
static void errdump(char *id, int err, int the_errno) {
	fprintf(stderr,"Error at %s : %d -> %s ; errno %d -> %s\n",
		id, err, errorname(err),the_errno,strerror(the_errno) );
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
static void dumpdevdesc( struct libusb_device_descriptor *desc) {
	fprintf(stderr,"Device Descriptor\n");
	fprintf(stderr," bLength 0x%x type 0x%x USB ver 0x%x\n",
	0 + desc -> bLength,
	0 + desc -> bDescriptorType,
	0 + desc -> 	bcdUSB );
	fprintf(stderr," Class 0x%x SubClass 0x%x Protocol 0x%x\n",
	0 + desc -> 	bDeviceClass,
	0 + desc -> 	bDeviceSubClass,
	0 + desc -> 	bDeviceProtocol );
	fprintf(stderr," MaxPacketSize0 0x%x\n",
	0 + desc -> 	bMaxPacketSize0 );
	fprintf(stderr," Vendor 0x%x Product 0x%x Device 0x%x\n",
	0 + desc -> 	idVendor,
	0 + desc -> 	idProduct,
	0 + desc -> 	bcdDevice );
	fprintf(stderr," iMfr 0x%x iProd 0x%x iSerNo 0x%x\n",
	0 + desc -> 	iManufacturer ,
	0 + desc -> 	iProduct ,
	0 + desc -> 	iSerialNumber );
	fprintf(stderr," NumConfig 0x%x\n",
	0 + desc -> 	bNumConfigurations );
}
static void decode_attribs(uint8_t attribs) {
	uint8_t v1;
	v1 = attribs&3;
	fprintf(stderr,"Attribs 0x%2.2x :",0+attribs);
		fprintf(stderr," Type ");
	switch(v1) {
	case LIBUSB_TRANSFER_TYPE_CONTROL :
		fprintf(stderr,"CTRL");
		break;
	case LIBUSB_TRANSFER_TYPE_ISOCHRONOUS :
		fprintf(stderr,"ISOC");
		break;
	case LIBUSB_TRANSFER_TYPE_BULK :
		fprintf(stderr,"BULK");
		break;
	case LIBUSB_TRANSFER_TYPE_INTERRUPT :
		fprintf(stderr,"INTR");
		break;
	default:
		fprintf(stderr,"????");
		break;
	}
	fprintf(stderr," Sync ");
	v1 = attribs&0xc;
	switch(v1) {
	case LIBUSB_ISO_SYNC_TYPE_NONE :
		fprintf(stderr,"NONE");
		break;
	case LIBUSB_ISO_SYNC_TYPE_ASYNC :
		fprintf(stderr,"ASYN");
		break;
	case LIBUSB_ISO_SYNC_TYPE_ADAPTIVE :
		fprintf(stderr,"ADAP");
		break;
	case LIBUSB_ISO_SYNC_TYPE_SYNC :
		fprintf(stderr,"SYNC");
		break;
	default:
		fprintf(stderr,"????");
		break;
	}
	fprintf(stderr," Usage ");
	v1 = attribs&0x30;
	switch(v1) {
	case LIBUSB_ISO_USAGE_TYPE_DATA :
		fprintf(stderr,"DATA");
		break;
	case LIBUSB_ISO_USAGE_TYPE_FEEDBACK :
		fprintf(stderr,"FDBK");
		break;
	case LIBUSB_ISO_USAGE_TYPE_IMPLICIT :
		fprintf(stderr,"IMPL");
		break;
	default:
		fprintf(stderr,"????");
		break;
	}
	fprintf(stderr," Reserved ");
	v1 = attribs&0xc0;
	fprintf(stderr,"0x%2.2x",0+v1);

	fprintf(stderr,"\n");
}
static void dumpepdesc(const struct libusb_endpoint_descriptor *endp, int num) {
	static char dir_in[] ="IN";
	static char dir_out[]="OUT";
	static char dir_bad[]="BAD";
	static char *dir_val;
	if( (0x80 & endp->bEndpointAddress) == LIBUSB_ENDPOINT_IN) dir_val = dir_in;
	else if( (0x80 & endp->bEndpointAddress) == LIBUSB_ENDPOINT_OUT) dir_val = dir_out;
	else dir_val = dir_bad;

	fprintf(stderr,"Endpoint %d Len 0x%x DescType 0x%x\n",
	num,
	0 + endp->bLength,
	0 + endp->bDescriptorType );
	fprintf(stderr,"Addr 0x%x %3s Attrib 0x%x PktSiz 0x%x\n",
	0 + endp->bEndpointAddress, dir_val,
	0 + endp->bmAttributes,
	0 + endp->wMaxPacketSize );
	fprintf(stderr,"Interval 0x%x Refresh 0x%x SyncAccr 0x%x\n",
	0 + endp->bInterval,
	0 + endp->bRefresh,
	0 + endp->bSynchAddress );
	// const unsigned char *extra
	fprintf(stderr,"Extra Len 0x%x\n",
	0 + endp->extra_length );
	decode_attribs(endp->bmAttributes);
	if(endp->extra_length != 0) {
		fprintf(stderr,"Extra TBD\n");
	}
}
static void dumpinterfacedesc(const struct libusb_interface_descriptor * ifp, int num, int alt ) {
	fprintf(stderr,"Interface %d Alt %d Len 0x%x Type 0x%x Number 0x%x Alt 0x%x\n",
	num, alt,
	0 + ifp->bLength,
	0 + ifp->bDescriptorType,
	0 + ifp->bInterfaceNumber,
	0 + ifp->bAlternateSetting );
	fprintf(stderr,"Endpoints 0x%x Class 0x%x SubClass 0x%x Protocol 0x%x\n",
	0 + ifp->bNumEndpoints,
	0 + ifp->bInterfaceClass,
	0 + ifp->bInterfaceSubClass,
	0 + ifp->bInterfaceProtocol );
	fprintf(stderr,"iInterface 0x%x extra_len 0x%x\n",
	0 + ifp->iInterface,
	0 + ifp->extra_length );
	int k;
	for(k=0;k<ifp->bNumEndpoints;k++) {
		dumpepdesc(ifp->endpoint+k,k);
	}
	if(ifp->extra_length != 0) {
		fprintf(stderr,"Extra TBD\n");
		// const unsigned char * 	extra
	}
}
static void dumpconfigdesc( struct libusb_config_descriptor *cfg, int num) {
	fprintf(stderr,"Config %d descriptor. Length 0x%x Type 0x%x Total Len 0x%x\n",
	num,
	0 + cfg->bLength,
	0 + cfg->bDescriptorType,
	0 + cfg->wTotalLength );
	fprintf(stderr,"NumInt 0x%x ConfigValue 0x%x iConfig 0x%x Attribs 0x%x MaxPwr 0x%x\n",
	0 + cfg->bNumInterfaces,
	0 + cfg->bConfigurationValue,
	0 + cfg->iConfiguration,
	0 + cfg->bmAttributes,
	0 + cfg->MaxPower );
	fprintf(stderr,"Extra Len 0x%x\n",
	0 + cfg->extra_length );
	int k;
	for(k=0;k<cfg->bNumInterfaces;k++) {
		const struct libusb_interface_descriptor * ifp;
		int p;
		p = cfg->interface[k].num_altsetting;
		ifp = cfg->interface[k].altsetting;
		dumpinterfacedesc(ifp,k,p);
	}
	if(cfg->extra_length != 0 ) {
		fprintf(stderr,"Extra TBD\n");
		// const unsigned char * 	extra
	}
}
static void dumpinfo(libusb_device *usbdev) {
	int err;
	int k;
	struct libusb_device_descriptor desc ;
	err = libusb_get_device_descriptor (usbdev, &desc);
	if(err != 0) errdump("dumpinfo_descriptor",err,errno);
	else dumpdevdesc(&desc);

	for(k=0;k<desc.bNumConfigurations;k++) {
		struct libusb_config_descriptor *config;
		err = libusb_get_config_descriptor (usbdev, k, &config);
		if(err != 0) errdump("dumpinfo_config",err,errno);
		else dumpconfigdesc(config,k);
	}

}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
// If chk is true exit on error, else return NULL.
static libusb_device *find_usb_device(libusb_context *context, uint16_t vend, uint16_t prod, int chk) {
	/******************************
	** Find a useful device. */
	libusb_device **usbdevlist;
	libusb_device *found = NULL;
	// fprintf(stdout,"F1\n"); fflush(stdout);
	ssize_t cnt = libusb_get_device_list(context, &usbdevlist);
	ssize_t i = 0;
	if (cnt < 0) {
		fprintf(stderr,"Trouble get_device_list %ld\n",cnt);
		if(chk) exit(2);
		else return found;
	}

	for (i = 0; i < cnt; i++) {
		fprintf(stderr,"Checking device %ld of %ld\n",i,cnt);
		// libusb_device *device = usbdevlist[i];
		struct libusb_device_descriptor desc;
		int r = libusb_get_device_descriptor(usbdevlist[i], &desc);
		if (r < 0) {
			fprintf(stderr, "failed to get device descriptor");
			if(chk) exit(2);
			else return found;
		} else {
			fprintf(stderr,"%04x:%04x (bus %d, device %d)\n",
				desc.idVendor, desc.idProduct,
				libusb_get_bus_number(usbdevlist[i]),
				libusb_get_device_address(usbdevlist[i]));
		}
		if ( (desc.idVendor==vend) && (desc.idProduct==prod) ) {
			found = usbdevlist[i];
			break;
		}
	}
	if(!found) {
		fprintf(stderr,"Device vendor 0x%4.4hx prod 0x%4.4hx not found\n",vend,prod);
		if(chk) exit(2);
	}
	// libusb_free_device_list(usbdevlist, 1);
	return found;
}
static void *sharedmem_addr = NULL;
static uint32_t *sharedmem_hdr = NULL;
static int sharedmem_hdr_len = 0;
static int8_t   *sharedmem_data5 = NULL;
static int sharedmem_data5_len = 0;
static int8_t   *sharedmem_data3 = NULL;
static int sharedmem_data3_len = 0;
static int8_t   *sharedmem_vals = NULL;
static int sharedmem_vals_len = 0;
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
static void setup_myshm(char *shmname) {
	int s_handle;
	int rtn;
	if(shmname == NULL) {
		fprintf(stderr,"No shared memory requested\n");
		return;
	}
	s_handle = shm_open(shmname,O_RDWR|O_CREAT,0644);
	if(s_handle<0) {
		fprintf(stderr,"Trouble calling shm_open %s\n",shmname);
		fprintf(stderr,"Error %d %s\n",errno,strerror(errno));
		exit(2);
	}
	rtn = ftruncate(s_handle,2048);
	if(rtn<0) {
		fprintf(stderr,"Trouble calling ftruncate %s\n",shmname);
		fprintf(stderr,"Error %d %s\n",errno,strerror(errno));
		exit(2);
	}
	sharedmem_addr = mmap(NULL,2048,PROT_READ|PROT_WRITE,MAP_SHARED,s_handle,0);
	if(sharedmem_addr == MAP_FAILED) {
		fprintf(stderr,"Trouble calling mmap2 %s\n",shmname);
		fprintf(stderr,"Error %d %s\n",errno,strerror(errno));
		exit(2);
	}
	close(s_handle); // Mapping stays active.
	sh_ptr = sharedmem_addr;
	sharedmem_hdr = sharedmem_addr;
	sharedmem_hdr_len = 4;
	sharedmem_data5 = ((void *)sharedmem_hdr) +sizeof(uint32_t)*sharedmem_hdr_len;
	sharedmem_data5_len = 64;
	sharedmem_data3 = ((void *)sharedmem_data5) +sizeof(int8_t)*sharedmem_data5_len;
	sharedmem_data3_len = 512;
	sharedmem_vals  = ((void *)sharedmem_data3) +sizeof(int8_t)*sharedmem_data3_len;
	sharedmem_vals_len = 512;
}
mqd_t mymq_desc = (mqd_t)(-1);
mqd_t mymqctl_desc = (mqd_t)(-1);
struct mq_attr mqattrs = {
  0,  // mq_flags
 10,  // mq_maxmsg
 64,  // mq_msgsize (8192 default)
  0,  // mq_curmsgs
};
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
static void setup_mymq(char *mqname, mqd_t *mqptr) {
	int rtn;
	struct mq_attr oldattrs;

	if(mqname == NULL) return;

	// Note: mqattrs is ignored if the queue already exists.
#ifdef NOTYET
	rtn = mq_unlink(mqname);
	if(rtn<0) {
		fprintf(stderr,"Trouble at mq_unlink %s\n",mqname);
		fprintf(stderr,"error %d <%s>\n",errno,strerror(errno));
	}
#endif
	*mqptr = mq_open(mqname,O_RDWR|O_CREAT,0644,&mqattrs);
	if(*mqptr == (mqd_t)(-1)) {
		fprintf(stderr,"Trouble at mq_open %s\n",mqname);
		fprintf(stderr,"error %d <%s>\n",errno,strerror(errno));
		exit(2);
	}
	rtn = mq_setattr(*mqptr,&mqattrs,&oldattrs);
	if(rtn<0) {
		fprintf(stderr,"Trouble at mq_setattr %s\n",mqname);
		fprintf(stderr,"error %d <%s>\n",errno,strerror(errno));
	}
	rtn = mq_getattr(*mqptr,&mqattrs);
	fprintf(stderr,"mq attrs flags 0x%lx maxmsg %ld size %ld cur %ld\n",
		mqattrs.mq_flags,mqattrs.mq_maxmsg,mqattrs.mq_msgsize,mqattrs.mq_curmsgs);
}
static void mymq_send(int val) {
	struct timespec tmout;
	int rtn;
	if(mymq_desc == (mqd_t)(-1)) return;
	tmout.tv_sec = 0; // timeout on Jan 1, 1970.
	tmout.tv_nsec = 0;
	rtn = mq_timedsend(mymq_desc,(char *)&val,sizeof(val),1, &tmout);
	if(rtn < 0) {
		fprintf(stderr,"Trouble at mq_timedsends\n");
		fprintf(stderr,"error %d <%s>\n",errno,strerror(errno));
	} else {
		fprintf(stderr,"Sent msg %d\n",val);
	}
}
static int mymq_check(mqd_t *thequ,uint32_t *val, int maxval) { // Returns number of ints if there's a message.
	ssize_t vsize;
	unsigned msg_pri;
	struct timespec tmout;
	char tmpd[100];
	tmout.tv_sec = 0; // timeout on Jan 1, 1970.
	tmout.tv_nsec = 0;
	int rtnval;
	if(*thequ == (mqd_t)(-1) ) {
		return 0; // No message if no queue.
	}
	rtnval = 0;
	vsize = mq_timedreceive(*thequ, tmpd, sizeof(tmpd),&msg_pri,&tmout);
	if(vsize<0) {
		if(errno==ETIMEDOUT) {
			// fprintf(stderr,"MQ timeout\n");
			// Do nothing for now.
		} else {
			fprintf(stderr,"Trouble at mq_timedreceive\n");
			fprintf(stderr,"error %d <%s>\n",errno,strerror(errno));
		}
	} else {
		int k;
		fprintf(stderr,"Got msg len %ld pri %u\n", vsize,msg_pri);
		rtnval = vsize/sizeof(*val);
		for(k=0;k<rtnval;k++) val[k] = ((int *)tmpd)[k];
	}
	return rtnval;
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
static void save_data(int8_t *indata, int len, int8_t *shrdata, int shrlen, int *ctrptr) {
	if(shrdata == NULL) return;
	if(len > shrlen) len = shrlen;
	memmove(shrdata,indata,len);
	memset(shrdata+len,0,shrlen-len);
	(*ctrptr)++;
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
// Registers:
// 0 read address
// 1 read count
// 2 cic factor - 1
// 3 cic output rightshift
// 4 DDS step.
// 5 strobes:
//   0: hilo_reset     (level; should be edge).
//   2: trigger enable (level; should be edge).
//   4: 1:rising edge, 0: falling edge.
//   8: Manual trigger (level; should be edge).
//  10: EP5 (USB interrupt channel) trigger
//  20: Set cic_write (start capturing; ends on post-trigger counter going to zero).
// 6 lag (number of samples captured after trigger)
// Initial setup:
int8_t outdata[64] =
// 12345678
  ":@=1@00="  // 8 Reg0 = 0; Reg1 = 0x100
  ":2@:13="   // 7 Reg2 = 19 (0x13) cic=20
  ":3@:e="    // 6 Reg3 = 14 (0xe) shift=14
  ":4@f4240=" // 9 Reg4=1000000 (0xf4240) dds=
  ":6@7d0="   // 7 Reg6=2000 (0x7d0) lag=
  ":5@29="    // 6 Reg5=0x29 - Reset; load post-trigger counter. This is a VHDL bug workaround.
  ":5@20="    // 6 Reg5=0x20 - Working mode.
;
int outdatalen = 49;
static void value2out( enum vhdl_regs reg, long factr) {
	int p;
	p=snprintf((char *)outdata,ArraySize(outdata),":%x@:%lx=   ",reg,factr);
	outdatalen=p-1;
	fprintf(stderr,"Sending %d <",p);
	fwrite(outdata,sizeof(*outdata),p,stderr);
	fprintf(stderr,">\n");
	sh_ptr->vals[reg] = factr;
}
static void string2out(enum vhdl_regs reg, char *cmdval) {
	long factr;
	factr=strtol(cmdval,0,0); // Allows octal 0 or hex 0x prefixes.
	value2out(reg,factr);
}
static int command_go(void) {
	strncpy((char *)outdata,":@=1@00=",8); // reg[0]=0; reg[1]=0x100; read address, count.
	outdata[8]='\0';
	outdatalen=8;
	return 1;
}
static int readmycommands(void) {
	char buf[101];
	int k;
	int rtnval;
	k = read(0,buf,100);
	rtnval = 0;
	if(k>0) {
		fprintf(stdout,"Got string k=%d <",k);
		fwrite(buf,sizeof(*buf),k,stdout);
		fprintf(stdout,">\n");
		buf[k]='\0'; // Just in case.
		if(strncmp(buf,"?",1)==0) {
			fprintf(stderr,"commands\n");
			fprintf(stderr,"? help\n");
			fprintf(stderr,"go\n");
			fprintf(stderr,"clr[=val]  Reset hilo, or set strobes\n");
          		fprintf(stderr," 0:hilo_reset 1:trigger_enable 2:trigger_on_rising"
          			"3:trigger_manual 4:ep5_trigger\n");

			fprintf(stderr,"cic=# Desampling factor.\n");
			fprintf(stderr,"shift=# Shift factor.\n");
			fprintf(stderr,"dds=# DDS step.\n");
			fprintf(stderr,"lag=# Samples after trigger.\n");
		} else if(strncmp(buf,"go",2)==0) {
			// strncpy((char *)outdata,":@=1@00=",8);
			// outdata[8]='\0';
			// outdatalen=8;
			// rtnval = 1;
			rtnval = command_go();
		} else if(strncmp(buf,"clr",3)==0) {
			if(buf[3]=='=') {
				string2out(VREG_FLAGS,buf+4);
			} else {
				strncpy((char *)outdata,":5@:1=   :=   ",13);
				outdata[13]='\0';
				outdatalen=13;
			}
			printf("sending %d <%s>\n",outdatalen,outdata);
			rtnval = 1;
		} else if(strncmp(buf,"cic=",4)==0) {
			int factr;
			factr=strtol(buf+4,0,0);
			value2out(VREG_CIC,factr>1?factr-1:0);
			rtnval = 1;
		} else if(strncmp(buf,"shift=",6)==0) {
			string2out(VREG_RTSHIFT,buf+6);
			rtnval = 1;
		} else if(strncmp(buf,"dds=",4)==0) {
			string2out(VREG_DDS,buf+4);
			rtnval = 1;
		} else if(strncmp(buf,"lag=",4)==0) {
			string2out(VREG_POSTTRIG,buf+4);
			rtnval = 1;
		} else {
			fprintf(stderr,"Unknown command \n");
			fwrite(buf,sizeof(*buf),k,stderr);
		}
	}
	return rtnval;
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
static int mymq_commands(int *mqctval, int num) {
	int rtnval;
	if( (num>=1)&&(mqctval[0]==10) ) { // Go command.
		command_go();
		rtnval = 1;
	} else if( (num>=2)&&(mqctval[0]==5) ) { // DDS value.
		value2out(VREG_DDS,mqctval[1]);
		rtnval = 1;
	} else if( (num>=2)&&(mqctval[0]==1) ) { // CIC factor.
		value2out(VREG_CIC,mqctval[1]>1?mqctval[1]-1:0);
		rtnval = 1;
	} else if( (num>=2)&&(mqctval[0]==2) ) { // shift.
		value2out(VREG_RTSHIFT,mqctval[1]);
		rtnval = 1;
	} else if( (num>=2)&&(mqctval[0]==11) ) { // flags.
		value2out(VREG_FLAGS,mqctval[1]);
		rtnval = 1;
	} else {
		int k;
		fprintf(stderr,"%s : unimplemented msg",__func__);
		for(k=0;k<num;k++) fprintf(stderr," 0x%x",mqctval[k]);
		fprintf(stderr,"\n");
		rtnval = 0;
	}
	return rtnval;
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
int main(int argc, char *argv[]) {
	int rtn;
	libusb_context *context;
	int ofile = -1;
	int k;
	int intf_num = 0;
	int intf_alt = 0;
	uint8_t ep;
	char *shmname = NULL;
	char *mqname = NULL;
	char *mqctlname = NULL;

	/* Change stdin to nonblocking mode. */

	int flags;
	flags = fcntl(0, F_GETFL); 
	flags |= O_NONBLOCK;	
	fcntl(0, F_SETFL, flags);

	for(k=1;k<argc;k++) {
		if(strncmp("-o",argv[k],2)==0) {
			ofile=open(argv[k]+2,O_WRONLY|O_CREAT|O_EXCL,0644);
			if(ofile < 0) {
				fprintf(stderr,"Trouble creating output file <%s>\n",argv[k]+2);
				exit(2);
			}
		} else if(strncmp("-e",argv[k],2)==0) {
			ep = strtoul(argv[k]+2,0,0);
		} else if(strncmp("-m",argv[k],2)==0) {
			shmname = argv[k]+2;
		} else if(strncmp("-q",argv[k],2)==0) {
			mqname = argv[k]+2;
		} else if(strncmp("-c",argv[k],2)==0) {
			mqctlname = argv[k]+2;
		} else {
			fprintf(stderr,"Bad argument <%s>\n",argv[k]);
			exit(2);
		};
	}
	if(ofile<0) {
		fprintf(stderr,"No output file\n");
		sleep(1);
	}
	// fprintf(stdout,"A\n"); fflush(stdout);

	signal(SIGINT,mysigint);

	setup_myshm(shmname);
	setup_mymq(mqname,&mymq_desc);
	setup_mymq(mqctlname,&mymqctl_desc);
	// fprintf(stdout,"B\n"); fflush(stdout);

	/******************************
	** Init libusb-1.0.          */

	rtn = libusb_init (&context);
	if(rtn != 0) {
		fprintf(stderr,"%s: init trouble %d\n",argv[0],rtn);
		exit(2);
	}
	libusb_set_debug(context,3); // Info to stdout, warn/error to stderr.
	// fprintf(stdout,"C\n"); fflush(stdout);

	/******************************
	** Find a useful device. */

	libusb_device *found = NULL;
	found = find_usb_device(context,0x00fe,0x2100,1) ;
	if (!found) {
		fprintf(stderr,"Device not found\n");
		exit(2);
	} else {
		fprintf(stderr,"Found it\n");
	}
	// fprintf(stdout,"D\n"); fflush(stdout);

	libusb_device_handle *handle = NULL;
	// fprintf(stdout,"E\n"); fflush(stdout);
	if (found) {
		int err;

		fprintf(stderr,"Opening\n");
		err = libusb_open(found, &handle);
		if (err) {
			fprintf(stderr,"Trouble opening device %d %s\n",err,errorname(err));
			exit(2);
		}
	// fprintf(stdout,"F1\n"); fflush(stdout);
		err = libusb_kernel_driver_active(handle,intf_num);
		if(err == 0) {
			fprintf(stderr,"No kernel driver active.\n");
		} else if(err == 1) {
			fprintf(stderr,"Releasing kernel driver.\n");
			err = libusb_detach_kernel_driver(handle,intf_num);
			if(err == 0) {
				fprintf(stderr,"kernel driver successfully detached\n");
				do_reattach = 1;
			} else {
				fprintf(stderr,"Error code %d %s detaching driver\n",err,errorname(err));
				exit(2);
			}
		} else {
			fprintf(stderr,"error %d %s from libusb_kernel_driver_active\n",err,errorname(err));
			exit(2);
		}
	// fprintf(stdout,"F2\n"); fflush(stdout);

		fprintf(stderr,"Claiming\n");
		err = libusb_claim_interface(handle,intf_num);
		if (err) {
			fprintf(stderr,"Trouble claim_interface %d %s\n",err,errorname(err));
			exit(2);
		}
	// fprintf(stdout,"F3\n"); fflush(stdout);
		dumpinfo(found);
		fprintf(stderr,"Alt %d\n",intf_alt);
		err = libusb_set_interface_alt_setting (handle,intf_num,intf_alt);
		if (err) {
			fprintf(stderr,"Trouble alt_setting %d %s\n",err,errorname(err));
			exit(2);
		}
	// fprintf(stdout,"F4\n"); fflush(stdout);
		fprintf(stderr,"Xfer 0x%2.2x\n",ep);
		// continuous_transfer(handle,ofile,ep);
		int loopcount;
		for(loopcount=0;!exitflag;loopcount++) {
			int8_t indata[512];
			static int msgflag = 1;
// Command characters are:
// 0-9,a-f - hex bytes
// : - Clear accum
// @ - Copy accum to addr
// = - Copy accum to data
// Addresses are:
// 0 - Read address
// 1 - Read length
// 2 - Decimation
// 3 - shift.
			// This is just a test.
			// int8_t outdata[] = ":123456=@            ";
			// This partly works: Starts continuous send.
			//int8_t outdata[] = ": @ = :1 @ :100 = :  "; // 0<=0 1<=0x100 (Read 256 from 0)
			// int8_t outdata[] = ":@=:1@:100=:7@:=         "; // Not working.
			int actlen;
			int timeout_ms = 100;
			// Main data channel.
			checked_bulk_transfer (handle,0x83,(char *)indata,512,&actlen,timeout_ms,1);
			if(actlen>0) {
				save_data(indata,actlen,sharedmem_data3,sharedmem_data3_len,((int *)sharedmem_hdr)+1);
				smalldump((uint8_t *)indata,actlen);
				mymq_send(sharedmem_hdr[1]);
			}
			// Int channel.
			checked_bulk_transfer (handle,0x85,(char *)indata,512,&actlen,timeout_ms,1);
			if(actlen>0) {
				save_data(indata,actlen,sharedmem_data5,sharedmem_data5_len,((int *)sharedmem_hdr)+2);
				if(1) smalldump((unsigned char *)indata,actlen);
				mymq_send(sharedmem_hdr[1]);
			}
			// Out control channel.
			if(msgflag) {
				fprintf(stdout,"REQ\n");
				checked_bulk_transfer (handle,0x04,(char *)outdata,outdatalen,&actlen,timeout_ms,0x23);
				msgflag = 0;
			}
			if( (loopcount%1000) == 0) {
				fprintf(stdout,"Loop %d\n",loopcount);
			}
			// New commands can come via the keyboard or messages.
			if(msgflag==0) msgflag=readmycommands(); // Check keyboard first.
			if(msgflag==0) {
				uint32_t mqctlval[2];
				int num;
				num=mymq_check(&mymqctl_desc,mqctlval,2);
				if(num>0) {
					fprintf(stderr,"Got msg");
					for(k=0;k<num;k++) fprintf(stderr," %d",mqctlval[k]);
					fprintf(stderr,"\n");
					msgflag = mymq_commands((int *)mqctlval,num);
				}
			}
		}
	} else {
		fprintf(stderr,"Not found.\n");
	}
	if(do_reattach) {
		int err;
		err = libusb_attach_kernel_driver(handle,intf_num);
		if(err == 0) {
			fprintf(stderr,"kernel driver reattached\n");
		} else {
			fprintf(stderr,"Error %d %s on kernel driver reattach\n",err,errorname(err));
		}
	} else {
		fprintf(stderr,"No driver to reattach\n");
	}

	/******************************
	** Done. */
	if(handle != NULL) libusb_close(handle);
	libusb_exit (context);
	return 0;
}
